#include "paperViewer.h"
#include <math.h>
#include <qimage.h>
#include <qfiledialog.h>

using namespace std;

// The column grid has 2*nb+1 columns on its side
static const int nb = 2;

// Draws a spiral
void Viewer::draw()
{
  // Subjective distance ('D' in the paper) is set so that the camera revolveAroundPoint
  // does not 'slide' on the paper. See mouse bindings ('H' key) to know how to change revolveAroundPoint.
  float subjDist = (camera()->position() - camera()->revolveAroundPoint()).norm();
  // Prevent division by zero
  if (fabs(subjDist) < 1.0E-2)
    subjDist = (subjDist > 0.0) ? 1.0E-2 : -1.0E-2;
  dynCan_.setSubjectiveDistance(subjDist);
  // A constant value may also give good results, depending on your application :
  // dynCan_.setSubjectiveDistance( 3.0 );
  
  // Texture composition must be done first
  dynCan_.updateParameters(*camera());

  // Draw the column field
  glShadeModel(GL_FLAT);
  for (int i=-nb; i<=nb; ++i)
    for (int j=-nb; j<=nb; ++j)
      {
	glPushMatrix();
	glTranslatef(i, j, 0.0); 
	glCallList(cylinderDL_);
	glPopMatrix();
      }
  
  // Draw paper texture
  dynCan_.drawPaper();
}

void Viewer::keyPressEvent(QKeyEvent *e)
{
  const Qt::ButtonState state = (Qt::ButtonState)(e->state() & Qt::KeyButtonMask);
  switch (e->key())
    {
    case Qt::Key_S :
      dynCan_.setWarped(! dynCan_.isWarped());
      updateGL();
      break;
	
    case Qt::Key_L :
      loadPaperTexture();
      updateGL();
      break;
	
    case Qt::Key_Left :
      if ((state == Qt::NoButton) || (state == Qt::Keypad))
	{
	  dynCan_.setFrequencyBase(dynCan_.frequencyBase() / 1.2);
	  updateGL();
	}
      break;
	
    case Qt::Key_Right :
      if ((state == Qt::NoButton) || (state == Qt::Keypad))
	{
	  dynCan_.setFrequencyBase(dynCan_.frequencyBase() * 1.2);
	  updateGL();
	}
      break;

    case Qt::Key_Up :
      if ((state == Qt::NoButton) || (state == Qt::Keypad))
	{
	  dynCan_.makePaperMoreTransparent();
	  updateGL();
	}
      break;
	
    case Qt::Key_Down :
      if ((state == Qt::NoButton) || (state == Qt::Keypad))
	{
	  dynCan_.makePaperLessTransparent();
	  updateGL();
	}
      break;
      
    default :
      QGLViewer::keyPressEvent(e);
      break;
    }
}

void Viewer::init()
{
  createCylinderDisplayList();

  setMouseBinding(Qt::ShiftButton | Qt::LeftButton, RAP_FROM_PIXEL);
  
  // Position camera
  camera()->setPosition(0.0, -3.0*nb, 1.4);
  camera()->setUpVector(0.0, 0.0, 1.0);
  // camera()->setPosition(0.0, 0.0, -3.0*nb);
  // camera()->setUpVector(0.0, 1.0, 0.0);
  camera()->lookAt(0.0, 0.0, 0.0);

  // Set camera in FLY mode
  toggleCameraMode();

  // Define scene size and activate XY grid display
  setSceneRadius(1.2*nb);
  setDrawGrid();

  dynCan_.initGL(width(), height(), *camera());
  dynCan_.loadPaperTexture("paper.jpg");
}

void Viewer::loadPaperTexture()
{
  QString name = QFileDialog::getOpenFileName(".", "Images (*.png *.xpm *.jpg)", this, "Choose", "Select a paper texture");
  
  // In case of Cancel
  if (name.isEmpty())
    return;

  QImage img(name);
  
  if (img.isNull())
    {
      cerr << "Unable to load " << name << ", unsupported file format" << endl;
      return;
    }

  cout << "Loading " << name << ", " << img.width() << "x" << img.height() << " pixels" << endl;

  // 1E-3 needed. Just try with width=128 and see !
  int newWidth  = 1<<(int)(1+log(img.width() -1+1E-3) / log(2.0));
  int newHeight = 1<<(int)(1+log(img.height()-1+1E-3) / log(2.0));
  
  if ((img.width()!=newWidth) || (img.height()!=newHeight))
    {
      cout << "Image size is not a power of two - Aborting" << endl;
      return;
    }
  
  dynCan_.loadPaperTexture(name);
}

QString Viewer::helpString() const
{
  QString text("<h2>D y n a m i c C a n v a s</h2>");
  text += "Implementation of the GI'03 DynamicCanvas paper.<br><br>";
  text += "Left (resp. right) button moves forward (resp. backward).<br>";
  text += "Middle button looks around without moving.<br>";
  text += "Shift + left click on an object determines the subjective distance.<br><br>";
  text += "Press '<b>S</b>' to toggle spherical warping.<br>";
  text += "Press '<b>L</b>' to load a new paper texture.<br>";
  text += "Left and right arrows change paper texture frequency.<br>";
  text += "Up and down arrows change paper texture transparency.<br>";
  text += "+/- change camera fly speed.<br>";
  return text;
}

void Viewer::createCylinderDisplayList()
{
  const float radius = 0.2;
  const float length = 1.0;
  const int nbSub = 8;
    
  cylinderDL_ = glGenLists(1);
  glNewList(cylinderDL_, GL_COMPILE);
  
  float prevX=radius, prevY=0.0f;
  float nx=1.0f, ny=0.0f;

  // Cylinder faces
  glBegin(GL_QUAD_STRIP);
  for (unsigned short i=1; i<=nbSub; ++i)
    {
      if (i%2)
	glColor3f(0.1, 0.1, 0.1);
      else
	glColor3f(0.8, 0.8, 0.8);
      
      glNormal3f(nx, ny, 0.0);
      glVertex3f(prevX , prevY , length);
      glNormal3f(nx, ny, 0.0);
      glVertex3f(prevX , prevY , 0.0);

      nx = cos(2.0*M_PI*i/static_cast<float>(nbSub));
      ny = sin(2.0*M_PI*i/static_cast<float>(nbSub));

      prevX = radius * nx;
      prevY = radius * ny;

      if (i%2)
	glColor3f(0.8, 0.8, 0.8);
      else
	glColor3f(0.1, 0.1, 0.1);

      glNormal3f(nx, ny, 0.0);
      glVertex3f(prevX , prevY , length);
      glNormal3f(nx, ny, 0.0);
      glVertex3f(prevX , prevY , 0.0);
    }
  glEnd();

  // Cylinder top
  glNormal3f(0.0, 0.0, 1.0);
  glBegin(GL_TRIANGLE_FAN);
  glVertex3f(0.0, 0.0, length);
  for (unsigned short i=0; i<=nbSub; ++i)
    {
      if (i%2)
	glColor3f(0.8, 0.8, 0.8);
      else
	glColor3f(0.1, 0.1, 0.1);

      prevX = radius * cos(2.0*M_PI*i/static_cast<float>(nbSub));
      prevY = radius * sin(2.0*M_PI*i/static_cast<float>(nbSub));

      glVertex3f(prevX , prevY , length);
    }
  glEnd();

  // Cylinder bottom
  glNormal3f(0.0, 0.0, -1.0);
  glBegin(GL_TRIANGLE_FAN);
  glVertex3f(0.0, 0.0, 0.0);
  for (unsigned short i=0; i<=nbSub; ++i)
    {
      if (i%2)
	glColor3f(0.8, 0.8, 0.8);
      else
	glColor3f(0.1, 0.1, 0.1);

      prevX = radius * cos(2.0*M_PI*i/static_cast<float>(nbSub));
      prevY = radius * sin(2.0*M_PI*i/static_cast<float>(nbSub));

      glVertex3f(prevX , prevY , 0.0);
    }
  glEnd();

  glEndList();
}
  
